@@ -1,8 +1,10 @@
 #!/usr/bin/python2
+# -*- coding: utf-8 -*-
 #
 # Automatic patch dependency checker and apply script/README.md generator.
 #
 # Copyright (C) 2014-2015 Sebastian Lackner
+# Copyright (C) 2015 Michael Müller
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -38,8 +40,8 @@ import subprocess
 import sys
 import tempfile
 import textwrap
-import urllib
-import xml.dom.minidom
+import xmlrpclib
+import ConfigParser
 
 _devnull = open(os.devnull, 'wb')
 
@@ -50,6 +52,7 @@ cached_patch_result    = {}
 
 class config(object):
     path_cache              = ".patchupdate.cache"
+    path_config             = os.path.expanduser("~/.config/patchupdate.conf")
 
     path_patches            = "patches"
     path_changelog          = "debian/changelog"
@@ -63,6 +66,14 @@ class config(object):
 
     path_IfDefined          = "9999-IfDefined.patch"
 
+    bugtracker_url          = "https://bugs.winehq.org/xmlrpc.cgi"
+    bugtracker_defaultcc    = ["michael@fds-team.de", "sebastian@fds-team.de",
+                               "erich.e.hoover@wine-staging.com"]
+    bugtracker_user         = None
+    bugtracker_pass         = None
+
+    github_url              = "https://github.com/wine-compholio/wine-staging"
+
 class PatchUpdaterError(RuntimeError):
     """Failed to update patches."""
     pass
@@ -145,34 +156,6 @@ def _parse_int(val, default=0):
     except AttributeError:
         return default
 
-def _winebugs_query(bugids):
-    """Query short_desc from multiple wine bugzilla bugs at the same time."""
-    bugids = list(set(bugids)) # Remove duplicates and convert to fixed order
-    if len(bugids) == 0: return {}
-
-    # Query bugzilla
-    url = "http://bugs.winehq.org/show_bug.cgi?%s&ctype=xml&field=short_desc&field=bug_status&field=resolution" % \
-          "&".join(["id=%d" % bugid for bugid in bugids])
-    with contextlib.closing(urllib.urlopen(url)) as fp:
-        data = xml.dom.minidom.parseString(fp.read())
-
-    # convert xml in a dictionary containing all bugs we found
-    result = {}
-    for element in data.getElementsByTagName('bug_id'):
-        bugids.remove(int(element.firstChild.data))
-
-    def _get_text(n):
-        return n.firstChild.data if (n and n.firstChild) else None
-
-    for bugid, element in zip(bugids, data.getElementsByTagName('bug')):
-        if element.getElementsByTagName('bug_id'): continue
-        result[bugid] = {
-            'short_desc': _get_text(element.getElementsByTagName('short_desc')[0]),
-            'bug_status': _get_text(element.getElementsByTagName('bug_status')[0]),
-            'resolution': _get_text(element.getElementsByTagName('resolution')[0]) }
-
-    return result
-
 def _read_changelog():
     """Read information from changelog."""
     with open(config.path_changelog) as fp:
@@ -180,12 +163,12 @@ def _read_changelog():
             r = re.match("^([a-zA-Z0-9][^(]*)\((.*)\) ([^;]*)", line)
             if r: yield (r.group(1).strip(), r.group(2).strip(), r.group(3).strip())
 
-def _latest_staging_version(only_stable=False):
-    """Get version number of the latest stable release."""
+def _latest_staging_version(only_released=False):
+    """Get version number of the latest release / unreleased version."""
     for package, version, distro in _read_changelog():
         if distro.lower() != "unreleased":
             return version
-        elif not only_stable:
+        elif not only_released:
             return "%s (unreleased)" % version
 
 def _latest_wine_commit(commit=None):
@@ -235,7 +218,6 @@ def read_patchset(revision = None):
     unique_id   = itertools.count()
     all_patches = {}
     name_to_id  = {}
-    all_bugids  = set()
     categories  = {}
 
     # Read in sorted order (to ensure created Makefile doesn't change too much)
@@ -313,19 +295,13 @@ def read_patchset(revision = None):
                 patch.categories.append(val)
 
             elif key == "fixes":
-                r = re.match("^[0-9]+$", val)
+                r = re.match("^\\[ *(!)? *([0-9]+) *\\](.*)$", val)
                 if r:
-                    bugid = int(val)
-                    patch.fixes.append((bugid, None))
-                    all_bugids.add(bugid)
+                    sync  = (r.group(1) != "!")
+                    bugid = int(r.group(2))
+                    patch.fixes.append((sync, bugid, r.group(3).strip()))
                     continue
-                r = re.match("^\\[ *([0-9]+) *\\](.*)$", val)
-                if r:
-                    bugid = int(r.group(1))
-                    patch.fixes.append((bugid, r.group(2).strip()))
-                    all_bugids.add(bugid)
-                    continue
-                patch.fixes.append((None, val))
+                patch.fixes.append((False, None, val))
 
             elif key == "disabled":
                 patch.disabled = _parse_int(val)
@@ -354,21 +330,6 @@ def read_patchset(revision = None):
         all_patches[i] = patch
         name_to_id[name] = i
 
-    # To simplify the task of keeping the bug list up-to-date, list all bugs
-    # which might require attention.
-    if revision is None:
-        once = True
-        for bugid, bug in sorted(_winebugs_query(all_bugids).items()):
-            if bug['bug_status'] not in ["UNCONFIRMED", "NEW", "ASSIGNED", "REOPENED"]:
-                if once:
-                    print ""
-                    print "WARNING: The following bugs might require attention:"
-                    print ""
-                    once = False
-                print " #%d - \"%s\" - %s %s" % (bugid, bug['short_desc'], bug['bug_status'],
-                                                 bug['resolution'] if bug['resolution'] else "")
-        print ""
-
     return all_patches
 
 def causal_time_combine(a, b):
@@ -471,6 +432,81 @@ def resolve_dependencies(all_patches, index = None, depends = None, auto_deps =
         _resolve(depends)
     return resolved
 
+def sync_bug_status(bugtracker, bug, url):
+    """Automatically updates the STAGED information of a referenced bug."""
+
+    # We don't want to reopen bugs
+    if bug['status'] not in ["UNCONFIRMED", "NEW", "ASSIGNED", "REOPENED", "STAGED"]:
+        return
+
+    if config.bugtracker_user is None or config.bugtracker_pass is None:
+        raise PatchUpdaterError("Can't update bug without username/password set")
+
+    changes = { 'ids'               : bug['id'],
+                'Bugzilla_login'    : config.bugtracker_user,
+                'Bugzilla_password' : config.bugtracker_pass }
+
+    # Update bug status
+    if bug['status'] != "STAGED":
+        changes['status'] = "STAGED"
+
+    # Update patchset URL
+    if bug['cf_staged_patchset'] != url:
+        changes['cf_staged_patchset'] = url
+
+    # Add missing CC contacts
+    missing_cc = []
+    for cc in config.bugtracker_defaultcc:
+        if cc not in bug['cc']:
+            missing_cc.append(cc)
+    if len(missing_cc):
+        changes["cc"] = {"add" : missing_cc}
+
+    bugtracker.Bug.update(changes)
+
+def check_bug_status(all_patches, sync_bugs=False):
+    """Checks the information in the referenced bugs and corrects them if sync_bugs is set."""
+
+    all_bugids = set()
+    url_map = {}
+
+    for _, patch in all_patches.iteritems():
+        url = "%s/tree/master/%s" % (config.github_url, patch.directory)
+        for sync, bugid, bugname in patch.fixes:
+            if sync and bugid is not None:
+                url_map[bugid] = url
+                all_bugids.add(bugid)
+
+    bugtracker  = xmlrpclib.ServerProxy(config.bugtracker_url)
+    bug_list    = bugtracker.Bug.get(dict(ids=list(all_bugids)))
+    staged_bugs = bugtracker.Bug.search(dict(status="STAGED"))
+
+    once = True
+    for bug in bug_list['bugs']:
+        if bug['status'] != "STAGED" or bug['cf_staged_patchset'] != url_map[bug['id']]:
+            if once:
+                print ""
+                print "WARNING: The following bugs might require attention:"
+                print ""
+                once = False
+            print " #%d - \"%s\" - %s %s - %s" % (bug['id'], bug['summary'], bug['status'],
+                                                  bug['resolution'], bug['cf_staged_patchset'])
+            if sync_bugs:
+                sync_bug_status(bugtracker, bug, url_map[bug['id']])
+
+    once = True
+    for bug in staged_bugs['bugs']:
+        if bug['id'] not in all_bugids:
+            if once:
+                print ""
+                print "WARNING: The following bugs are incorrectly marked as STAGED:"
+                print ""
+                once = False
+            print " #%d - \"%s\" - %s %s" % (bug['id'], bug['summary'], bug['status'],
+                                             bug['resolution'])
+
+    print ""
+
 def generate_ifdefined(all_patches, skip_checks=False):
     """Update autogenerated ifdefined patches, which can be used to selectively disable features at compile time."""
     enabled_patches = dict([(i, patch) for i, patch in all_patches.iteritems() if not patch.disabled])
@@ -615,22 +651,21 @@ def generate_script(all_patches, skip_checks=False):
                         for i, patch1 in set_apply:
                             for j, patch2 in set_skip:
                                 if causal_time_smaller(patch2.verify_time, patch1.verify_time):
-                                    return None # we can skip this test
+                                    return True # we can skip this test
 
                         try:
                             original = original_content
                             for i, _ in set_apply:
                                 original = patchutils.apply_patch(original, selected_patches[i][1], fuzz=0)
                         except patchutils.PatchApplyError:
-                            return current
+                            return False
 
-                        return None # everything is fine
+                        return True # everything is fine
 
                     def test_apply_seq(current_list):
                         for current in current_list:
-                            failed = test_apply(current)
-                            if failed is not None:
-                                return failed
+                            if not test_apply(current):
+                                return current
                         return None
 
                     iterables = []
@@ -647,6 +682,10 @@ def generate_script(all_patches, skip_checks=False):
                 # Update the dependency cache
                 dependency_cache[filename] = unique_hash
 
+            # Delete outdated cache information
+            for filename in dependency_cache.keys():
+                if not modified_files.has_key(filename):
+                    del dependency_cache[filename]
         finally:
             pool.close()
             _save_dict(config.path_cache, dependency_cache)
@@ -724,9 +763,9 @@ def generate_script(all_patches, skip_checks=False):
             lines.append("# |\n")
 
         # List all bugs fixed by this patchset
-        if any([bugid is not None for bugid, bugname in patch.fixes]):
+        if any([bugid is not None for sync, bugid, bugname in patch.fixes]):
             lines.append("# | This patchset fixes the following Wine bugs:\n")
-            for bugid, bugname in patch.fixes:
+            for sync, bugid, bugname in patch.fixes:
                 if bugid is not None:
                     lines.append("# |   *\t%s\n" % "\n# | \t".join(textwrap.wrap("[#%d] %s" % (bugid, bugname), 120)))
             lines.append("# |\n")
@@ -759,7 +798,7 @@ def generate_script(all_patches, skip_checks=False):
     # Add changes to git
     subprocess.call(["git", "add", config.path_script])
 
-def generate_markdown(all_patches, stable_patches):
+def generate_markdown(all_patches, release_patches):
     """Generate README.md including information about specific patches and bugfixes."""
 
     def _format_bug(mode, bugid, bugname):
@@ -772,13 +811,13 @@ def generate_markdown(all_patches, stable_patches):
 
     # Get fixes for current version
     for _, patch in all_patches.iteritems():
-        for bugid, bugname in patch.fixes:
+        for sync, bugid, bugname in patch.fixes:
             key = bugid if bugid is not None else bugname
             all_fixes[key] = [1, bugid, bugname]
 
-    # Compare with fixes for latest stable version
-    for _, patch in stable_patches.iteritems():
-        for bugid, bugname in patch.fixes:
+    # Compare with fixes for last release
+    for _, patch in release_patches.iteritems():
+        for sync, bugid, bugname in patch.fixes:
             if bugid is not None and all_fixes.has_key(bugid):
                 all_fixes[bugid][0] = 0
             elif all_fixes.has_key(bugname):
@@ -870,24 +909,39 @@ if __name__ == "__main__":
     parser = argparse.ArgumentParser(description="Automatic patch dependency checker and apply script/README.md generator.")
     parser.add_argument('--skip-checks', action='store_true', help="Skip dependency checks")
     parser.add_argument('--commit', type=_check_commit_hash, help="Use given commit hash instead of HEAD")
+    parser.add_argument('--sync-bugs', action='store_true', help="Update bugs in bugtracker (requires admin rights)")
     args = parser.parse_args()
 
     tools_directory = os.path.dirname(os.path.realpath(__file__))
     os.chdir(os.path.join(tools_directory, "./../.."))
 
+    config_parser = ConfigParser.ConfigParser()
+    config_parser.read(config.path_config)
+
+    try:
+        config.bugtracker_user = config_parser.get('bugtracker', 'username')
+        config.bugtracker_pass = config_parser.get('bugtracker', 'password')
+    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
+        config.bugtracker_user = None
+        config.bugtracker_pass = None
+
     try:
 
         # Get information about Wine and Staging version
         latest_wine_commit       = _latest_wine_commit(args.commit)
-        latest_staging_version   = _latest_staging_version(only_stable=True)
+        latest_staging_version   = _latest_staging_version(only_released=True)
 
-        # Read current and stable patches
-        all_patches    = read_patchset()
-        stable_patches = read_patchset(revision="v%s" % latest_staging_version)
+        # Read current and release patches
+        all_patches     = read_patchset()
+        release_patches = read_patchset(revision="v%s" % latest_staging_version)
 
+        # Check bugzilla
+        check_bug_status(all_patches, sync_bugs=args.sync_bugs)
+
+        # Update autogenerated files
         generate_ifdefined(all_patches, skip_checks=args.skip_checks)
         generate_script(all_patches, skip_checks=args.skip_checks)
-        generate_markdown(all_patches, stable_patches)
+        generate_markdown(all_patches, release_patches)
         wrap_changelog()
 
     except PatchUpdaterError as e:
